package org.semanticweb.triple.application;

import java.util.*;
import java.io.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.*;
import org.semanticweb.triple.parser.*;
import org.semanticweb.rdf.utilities.*;
import org.semanticweb.triple.language.*;
import org.semanticweb.glp.*;


public class TripleProgram {

  final static String namespace = "http://www.w3.org/2001/06/30/triple#";
  final static String pkg = "dfki.rdf.triple.language";

  Collection _objects; // of TripleExpression

  public TripleProgram() {
    _objects = new ArrayList();
  }

  public TripleProgram(String rdfFilename) throws Exception {
    _objects =
      (new RDFImport(namespace, pkg)).importObjects(rdfFilename).values();
  }

  public TripleProgram(LTerm program) throws Exception {
    _objects = new ArrayList();
    transformClauses(program);
  }

  public void addClause(Clause clause) {
    _objects.add(clause);
  }


  // test

  public static void main(String args[]) {
    System.out.println("TRIPLE running ...\n");
    // arg 1: rdf or triple file
    if (args.length >= 1) {
      String filename = args[0];
      try {
        // TripleProgram pgm = new TripleProgram(filename); // rdf
        LTermParser parser = null;
        try {
          parser = new LTermParser(new FileInputStream(filename));
        } catch (FileNotFoundException e) {
          System.out.println (e);
          System.exit(1);
        }
        LTerm lpgm = parser.parse();
        System.out.println("Hilog Term:");
        System.out.println(lpgm);
        TripleProgram pgm = new TripleProgram(lpgm);
        System.out.println("\nTRIPLE0/ASCII:\n");
        pgm.toASCII(new PrintWriter(System.out, true));
        System.out.println("\nHorn Logic:\n");
        pgm.toHornLogic(new PrintWriter(System.out, true), false);
        /* export test
        String exportFilename = rdfFilename + ".exported";
        (new RDFExport(namespace))
          .exportObjects(pgm._objects, exportFilename);
        System.out.println("\nExported to " + exportFilename + "\n");
        */
        ArrayList glp = pgm.toGLProgram(); // list of GLPClause
        System.out.println("\nGeneral Logic Program:");
        System.out.println(glp);
      } catch (Exception e) {
        e.printStackTrace();
      }

    } else {
      System.err.println("Usage: <triple file>");
    }
  }


  // output

  public void toASCII(PrintWriter out) throws Exception {

      // collect all models
      HashSet models = new HashSet();
      for (Iterator objIterator = _objects.iterator();
           objIterator.hasNext();) {
        Object object = objIterator.next();
        if (object instanceof Clause) {
          Term model = ((Clause)object).getModel();
          if (model != null)
            models.add(model);
        }
      }

      // show all models

      // clauses without model
      for (Iterator objIterator = _objects.iterator();
           objIterator.hasNext();) {
        Object object = objIterator.next();
        if (object instanceof Clause) {
          Term model = ((Clause)object).getModel();
          if (model == null)
            out.println(((Clause)object).toTripleString(false));
        }
      }

      // clauses with model
      for (Iterator modelIterator = models.iterator();
           modelIterator.hasNext();) {
        out.println();
        Term model = (Term)modelIterator.next();
        out.println("@" + model.toTripleString() + " {");
        for (Iterator objIterator = _objects.iterator();
             objIterator.hasNext();) {
          Object object = objIterator.next();
          if (object instanceof Clause) {
            Term cmodel = ((Clause)object).getModel();
            if (model == cmodel)
              out.println("  " + ((Clause)object).toTripleString(false));
          }
        }
        out.println("}");
      }

  }


  public void toHornLogic(PrintWriter out, boolean prologQuery)
      throws Exception {
    for (Iterator objIterator = _objects.iterator(); objIterator.hasNext();) {
      Object object = objIterator.next();
      if (object instanceof Clause) {
        Clause clause = (Clause)((Clause)object).toHornLogic();
        out.println(((Clause)clause).toString(true, false, prologQuery));
      }
    }
  }

  
  // TripleProgram -> GLP transformation  
  public ArrayList toGLProgram() { // returns list of GLPClause
    ArrayList glp = new ArrayList();
    Context context = new Context();
    for (Iterator objIterator = _objects.iterator(); objIterator.hasNext();) {
      TripleExpression expr = (TripleExpression)objIterator.next();      
      GLPExpression glpexpr = expr.toGLP(context);
      if (glpexpr instanceof GLPClause)
        glp.add(glpexpr);
    }
    return glp;
  }

  
  // -------------------------------------
  // LTerm -> TripleProgram transformation
  // -------------------------------------

  // transform clauses

  void transformClauses(LTerm program) throws ParseException {
    transformClauses(program, null); // no model
  }

  void transformClauses(LTerm lclause, Term model) throws ParseException {
    if (lclause instanceof LStruct) {
      if (isClause((LStruct)lclause)) {
        transformClauses(getSingleArg((LStruct)lclause), model);
      } else if (isMolecule((LStruct)lclause)) { // molecule
        addFacts(transformAtom((LStruct)lclause), model);
      } else if (isAndFormula((LStruct)lclause)) { // inner block w/o model
        transformClauses(getFirstArg((LStruct)lclause), model);
        transformClauses(getSecondArg((LStruct)lclause), model);
      } else if (isAtFormula((LStruct)lclause)) {
        // fact or inner block w/ model
        LTerm clauses = getFirstArg((LStruct)lclause);
        LTerm lmodel = getSecondArg((LStruct)lclause);
        Term model1 = transformTerm(lmodel, null); // no toplevel terms expected
        transformClauses(clauses, model1);
      } else if (isImpliedBy((LStruct)lclause)) {
        ArrayList args = ((LStruct)lclause).getArgs();
        if (args.size() == 1) {  // query
          LTerm lbody = getFirstArg((LStruct)lclause);
          transformQuery(lbody, model);
        } else if (args.size() == 2) {  // rule
          LTerm lhead = getFirstArg((LStruct)lclause);
          LTerm lbody = getSecondArg((LStruct)lclause);
          transformRule(lhead, lbody, model);
        } else {
          throw new ParseException("illegal <IMPLIEDBY>: " + lclause);
        }
      } else { // fact
        addFacts(transformAtom((LStruct)lclause), model);
      }
    } else if (lclause instanceof LSymbol) { // simple fact
      addFacts(transformAtom((LSymbol)lclause), model);
    } else {
      throw new ParseException("clause expected: " + lclause);
    }
  }

  void addFacts(Collection headAtoms, Term model) {
    for (Iterator atomIterator = headAtoms.iterator();
         atomIterator.hasNext();) {
      Clause clause = new Clause();
      clause.putHead((Atom)atomIterator.next());
      if (model != null)
        clause.putModel(model);
      _objects.add(clause);
    }
  }

  void transformQuery(LTerm lbody, Term model)
      throws ParseException {
    Formula body = transformBody(lbody);
    Clause clause = new Clause();
    clause.putBody(body);
    if (model != null)
      clause.putModel(model);
    _objects.add(clause);
  }

  void transformRule(LTerm lhead, LTerm lbody, Term model)
      throws ParseException {
    ArrayList headAtoms = transformAtom(lhead);
    Formula body = transformBody(lbody);
    for (Iterator headIterator = headAtoms.iterator();
         headIterator.hasNext();) {
      Atom headAtom = (Atom)headIterator.next();
      Clause clause = new Clause();
      clause.putHead(headAtom);
      clause.putBody(body);
      if (model != null)
        clause.putModel(model);
      _objects.add(clause);
    }
  }

  ArrayList transformAtom(LTerm latom) throws ParseException {
    // returns all atoms
    if (latom instanceof LStruct) {
      if (isMolecule((LStruct)latom)) {
        return transformMolecule((LStruct)latom);
      } else if (isAtFormula((LStruct)latom)) {
        LTerm latom1 = getFirstArg((LStruct)latom);
        LTerm lmodel = getSecondArg((LStruct)latom);
        ArrayList atoms = transformAtom(latom1);
        Term model = transformTerm(lmodel, null); // no toplevel terms expected
        for (Iterator atomIterator = atoms.iterator();
             atomIterator.hasNext();) {
          Atom atom = (Atom)atomIterator.next();
          if (atom.getModel() == null)
            atom.putModel(model);
        }
        return atoms;
      } else { // normal structure
        ArrayList atoms = new ArrayList();
        HornAtom atom = new HornAtom();
        String lfunctor = getFunctorSymbol((LStruct)latom);
        if (lfunctor != null) {
          atom.putPredicateSymbol(lfunctor);
          Collection largs = ((LStruct)latom).getArgs();
          for (Iterator largIterator = largs.iterator();
               largIterator.hasNext();) {
            atom.putArgs(transformTerm((LTerm)largIterator.next(), atoms));
          }
          atoms.add(atom);
          return atoms;
        } else {
          throw new ParseException("hilog not yet supported: " + latom);
        }
      }
    } else if (latom instanceof LSymbol) {
      ArrayList atoms = new ArrayList();
      HornAtom atom = new HornAtom();
      atom.putPredicateSymbol(((LSymbol)latom).getSymbol());
      atoms.add(atom);
      return atoms;
    } else {
      throw new ParseException("atom expected: " + latom);
    }
  }

  ArrayList transformMolecule(LStruct lmolecule) throws ParseException {
    // returns list of atoms
    ArrayList atoms = new ArrayList();
    LTerm lsubject = getSBFunctor(lmolecule); // [](s)(...) -> s
    // iterator over all pairs p->o
    for (Iterator pairIterator = lmolecule.getArgs().iterator();
         pairIterator.hasNext();) {
      LTerm pair = (LTerm)pairIterator.next();
      if (pair instanceof LStruct) {
        LTerm pfunctor = ((LStruct)pair).getFunctor();
        if (pfunctor instanceof LSymbol
            && isIMPLIES(((LSymbol)pfunctor).getSymbol())) {
          ArrayList args = ((LStruct)pair).getArgs();
          if (args.size() == 2) {
            Triple triple = new Triple();
            Iterator argsIterator = args.iterator();
            LTerm lpredicate = (LTerm)argsIterator.next();
            LTerm lobject = (LTerm)argsIterator.next();
            triple.putSubject(transformTerm(lsubject, atoms));
            triple.putPredicate(transformTerm(lpredicate, atoms));
            triple.putObject(transformTerm(lobject, atoms));
            atoms.add(triple);
          } else {
            throw new ParseException("molecule term expected: " + lmolecule);
          }
        } else {
          throw new ParseException("molecule term expected: " + lmolecule);
        }
      } else {
        throw new ParseException("molecule term expected: " + lmolecule);
      }
    }
    return atoms;
  }

  Formula transformBody(LTerm lbody) throws ParseException {
    if (lbody instanceof LStruct) {
      String functor = getFunctorSymbol((LStruct)lbody);
      if (isAndFormula((LStruct)lbody)) {
        LTerm firstTerm = getFirstArg((LStruct)lbody);
        LTerm secondTerm = getSecondArg((LStruct)lbody);
        And and = new And();
        and.putFirstFormula(transformBody(firstTerm));
        and.putSecondFormula(transformBody(secondTerm));
        return and;
      } // else isOR ... !!!
    }
    ArrayList atoms = transformAtom(lbody); // -> And
    if (atoms != null && atoms.size() > 0)
      return andFormula(atoms, 0);
    else
      throw new ParseException("illegal body: " + lbody);
  }

  Formula andFormula(ArrayList formulas, int i) throws ParseException {
    // formulas must be non-empty!
    if (i == formulas.size()-1)
      return (Formula)formulas.get(i);
    else {
      Formula restFormula = andFormula(formulas, i+1);
      And and = new And();
      and.putFirstFormula((Formula)formulas.get(i));
      and.putSecondFormula(restFormula);
      return and;
    }
  }


  // transform terms

  Term transformTerm(LTerm lterm, ArrayList atoms) throws ParseException {
    // in atoms, further top level atoms are collected (from molecules)
    if (lterm instanceof LSymbol)
      return new org.semanticweb.triple.language.Literal(((LSymbol)lterm).getSymbol());
    else if (lterm instanceof LVariable)
      return new Variable(((LVariable)lterm).getVariable());
    // else if ... numbers !!!
    else if (lterm instanceof LStruct) {
      if (isResource((LStruct)lterm)) {
        ArrayList args = ((LStruct)lterm).getArgs();
        Iterator argsIterator = args.iterator();
        // HACK:
        String uri =
          argsIterator.next().toString() + ":" + argsIterator.next();
        return new org.semanticweb.triple.language.Resource(uri);
      } else if (isMolecule((LStruct)lterm)) {
        return transformMoleculeTerm((LStruct)lterm, atoms);
      } else if (isAtFormula((LStruct)lterm)) {
        // ???
        LTerm lterm1 = getFirstArg((LStruct)lterm);
        LTerm lmodel = getSecondArg((LStruct)lterm);
        ArrayList atoms1 = new ArrayList();
        Term term1 = transformTerm(lterm1, atoms1);
        Term model = transformTerm(lmodel, null); // no toplevel terms expected
        for (Iterator atomIterator = atoms1.iterator();
             atomIterator.hasNext();) {
          Atom atom = (Atom)atomIterator.next();
          if (atom.getModel() == null)
            atom.putModel(model);
          atoms.add(atom);
        }
        return term1;
      } else { // normal term
        Structure struct = new Structure();
        String functor = getFunctorSymbol((LStruct)lterm);
        if (functor != null)
          struct.putFunctor(functor);
        else
          throw new ParseException("hilog not yet supported: " + lterm);
        Collection largs = ((LStruct)lterm).getArgs();
        for (Iterator largIterator = largs.iterator();
             largIterator.hasNext();) {
          struct.putArgs(transformTerm((LTerm)largIterator.next(), atoms));
        }
        return struct;
      }
    } else
      throw new ParseException("not yet implemented: " + lterm);
  }

  Term transformMoleculeTerm(LStruct sbterm, ArrayList atoms)
      throws ParseException {
    if (atoms != null)
      atoms.addAll(transformMolecule(sbterm));
    else
      throw new ParseException("molecule not expected: " + sbterm);
    ArrayList dummy = new ArrayList();
    return transformTerm(getSBFunctor(sbterm), dummy); // subject term
    // optimize: share subject term ... !!!
  }


  // auxiliaries

  String getFunctorSymbol(LStruct structure) {
    // returns null if functor is not a symbol
    LTerm functor = structure.getFunctor();
    if (functor instanceof LSymbol)
      return (((LSymbol)functor).getSymbol());
    else
      return null;
  }

  String getUltimateFunctorSymbol(LStruct structure) {
    LTerm functor = structure.getFunctor();
    if (functor instanceof LStruct)
      return getUltimateFunctorSymbol((LStruct)functor);
    else if (functor instanceof LSymbol)
      return (((LSymbol)functor).getSymbol());
    else
      return null;
  }

  boolean isMolecule(LStruct struct) {
    return getSBFunctor(struct) != null;
  }

  LTerm getSBFunctor(LStruct struct) { // [](f)(...) -> f
    // returns null if struct is not a square brackets structure
    LTerm functor = struct.getFunctor();
    if (functor instanceof LStruct) {
      LTerm ffunctor = ((LStruct)functor).getFunctor();
      if (ffunctor instanceof LSymbol
          && isSB(((LSymbol)ffunctor).getSymbol())) {
        ArrayList args = ((LStruct)functor).getArgs();
        if (args.size() == 1)
          return (LTerm)args.iterator().next();
        else
          return null;
      } else
        return null;
    } else
      return null;
  }

  LTerm getSingleArg(LStruct structure) throws ParseException {
    Collection args = structure.getArgs();
    if (args != null && args.size() == 1)
      return (LTerm)args.iterator().next();
    else
      throw new ParseException("one argument expected: " + structure);
  }

  LTerm getFirstArg(LStruct structure) throws ParseException {
    Collection args = structure.getArgs();
    if (args != null && args.size() > 0)
      return (LTerm)args.iterator().next();
    else
      throw new ParseException("argument expected: " + structure);
  }

  LTerm getSecondArg(LStruct structure) throws ParseException {
    Collection args = structure.getArgs();
    if (args != null && args.size() > 1) {
      Iterator argsIterator = args.iterator();
      argsIterator.next();
      return (LTerm)argsIterator.next();
    } else
      throw new ParseException("arguments expected: " + structure);
  }

  boolean isResource(LStruct struct) { // :(n,l)
    LTerm functor = struct.getFunctor();
    if (functor instanceof LSymbol
        && isCOLON(((LSymbol)functor).getSymbol())) {
      ArrayList args = struct.getArgs();
      return args.size() == 2;
    }
    return false;
  }

  boolean isImpliedBy(LStruct struct) { // [head] <- body
    LTerm functor = struct.getFunctor();
    if (functor instanceof LSymbol
        && isIMPLIEDBY(((LSymbol)functor).getSymbol())) {
      ArrayList args = struct.getArgs();
      return args.size() == 1 || args.size() == 2;
    }
    return false;
  }

  boolean isAtFormula(LStruct struct) { // @(f,m)
    LTerm functor = struct.getFunctor();
    if (functor instanceof LSymbol
        && isAT(((LSymbol)functor).getSymbol())) {
      ArrayList args = struct.getArgs();
      return args.size() == 2;
    }
    return false;
  }

  boolean isAndFormula(LStruct struct) { // f1 and f2
    LTerm functor = struct.getFunctor();
    if (functor instanceof LSymbol
        && isAND(((LSymbol)functor).getSymbol())) {
      ArrayList args = struct.getArgs();
      return args.size() == 2;
    }
    return false;
  }

  boolean isClause(LStruct struct) { // CLAUSE(clause)
    LTerm functor = struct.getFunctor();
    if (functor instanceof LSymbol
        && isCLAUSE(((LSymbol)functor).getSymbol())) {
      ArrayList args = struct.getArgs();
      return args.size() == 1;
    }
    return false;
  }


  // token tests

  boolean isSB(String functor) {
    return "[]".equals(functor);
  }

  boolean isCOLON(String functor) {
    return ":".equals(functor);
  }

  boolean isAT(String functor) {
    return "@".equals(functor);
  }

  boolean isIMPLIES(String functor) {
    return "->".equals(functor);
  }

  boolean isIMPLIEDBY(String functor) {
    return "<-".equals(functor)
      || ":-".equals(functor)
      || "?-".equals(functor);
  }

  boolean isAND(String functor) {
    return "AND".equals(functor)
      || "and".equals(functor)
      || ",".equals(functor);
  }

  boolean isCLAUSE(String functor) {
    return "CLAUSE".equals(functor)
      || "clause".equals(functor);
  }

  
}

